#!/usr/bin/env python
# -*- coding: utf-8 -*-

import bson, time, datetime
from pymongo import MongoClient
from datarest_db import DataRest_DB


class Service_DB():
    def __init__(self, message, mongo_ip, mongodb_port):
        # Conecta ao banco de dados
        self.mongo_server = MongoClient(mongo_ip, mongodb_port)

        self.__datarest = DataRest_DB(message)

        self.action()

    # --------------------------------
    # action
    # --------------------------------
    def action(self):
        action = self.__datarest.action()
        if hasattr(self, action):
            method = getattr(self, action)
            method()
        else:
            self.__datarest.add_error("ACTION ERRADA - MONGO DB", action + ":: action nao existe no servico")

    def send(self):
        return self.__datarest.send()

    # ----------------------------------------------------------------
    #
    # Public Methods
    #
    # ----------------------------------------------------------------
    # --------------------------------
    # find
    # --------------------------------
    # Realiza consulta na base de dados buscando um ou muitos documentos
    def find(self):
        # Recupera a coleção principal
        collection = self.__collection()

        # Recupera configurações
        filter = self.__datarest.filter()
        fields = self.__datarest.fields()
        limit = self.__datarest.limit()

        # Consulta o banco de dados
        find_result = collection.find(filter, fields)
        find_count = find_result.count()

        # Prepara configurações que serão retornadas para o cliente
        conditions = {
            "query": {
                "filter": filter, # TODO: Preparar filtro para ser retornado ao cliente
                "limit": limit
            },
            "count": find_count
        }

        # Obtém o resultado da consulta
        find_result = list(find_result.limit(limit))

        # Prepara os dados dos documentos recuperados $date, $timestamp ...
        find_result = self.__prepare_result(find_result)

        # Prepara os resultados para serem retornados ao cliente
        self.__datarest.result(find_result, conditions)


    # --------------------------------
    # insert
    # --------------------------------
    # Insere um ou muitos documentos na base de dados
    def insert(self):
        # Recupera a coleção principal
        collection = self.__collection()

        # Recupera dados e configurações
        data = self.__datarest.data()
        filter = self.__datarest.filter()
        fields = self.__datarest.fields()

        # Verifica o tipo de dados
        if type(data) == dict and len(data.keys()) > 0:
            data = [data]

        # Verifica se tem dados para inseretir na base de dados
        if len(data) > 0:
            # Insere documentos na base de dados e recupera a lista dos identificadores dos documentos inseridos
            insert_result = collection.insert_many(data).inserted_ids

            # Verifica se algum documento foi inserido
            if len(insert_result) > 0:
                # Prepara filtro para recuperar lista de documentos inseridos
                filter = {"_id": {"$in": insert_result}}

                # Realiza consulta na base de dados e recupera a lista de documentos inseridos
                find_result = collection.find(filter, fields)

                # Prepara os dados dos documentos recuperados $date, $timestamp ...
                find_result = self.__prepare_result(find_result)

                # Prepara os resultados para serem retornados ao cliente
                self.__datarest.result(find_result)
                return

        # Em case de erros por fata de documentos para inserir na base de dados ou qualquer outra falha no momento da inserção
        # registra mensagem de erro e configura a flha na operação
        self.__datarest.add_error(485752, "Não tem documentos para inserir na base de dados")

        # Registra resultado que será retornado para o servidor rest
        self.__datarest.result([], {
            "success": False
        })

    # --------------------------------
    # update
    # --------------------------------
    def update(self):
        # Recupera a coleção principal
        collection = self.__collection()
        update_count = 0

        # Recupera o filtro
        query_filter = self.__datarest.filter

        # TODO:: UPDATE - Verificar se existe filtro definido

        # Realiza busca na colecção para recuperar documentos que serão atualizados
        find_result = collection.find(query_filter)
        find_result_count = find_result.count()

        # Verifica se existe documentos para serem atualizados
        if find_result_count > 0:
            # Recupera a coleção changes
            collection_changes = self.__collection_changes()
            # Grava os documentos localizados na coleção trash
            insert_many_result = collection_changes.insert_many(find_result)

            # Verifica se os documentos foram realmente gravados na coleção trash
            if len(insert_many_result.inserted_ids) == find_result_count:
                # Recupera documento para salvar no banco de dados
                update_data = self.__datarest.data

                # Atualiza os documentos da coleção principal
                update_result = collection.update_many(query_filter, update_data, upsert=True)
                update_count = update_result.modified_count

        self.__datarest.result({"modified_count": update_count})

    # --------------------------------
    # remove
    # --------------------------------
    # Remove um ou muitos documentos na base de dados
    #
    def remove(self):
        # Recupera a coleção principal
        collection = self.__collection()
        deleted_count = 0

        # Recupera o filtro
        query_filter = self.__datarest.filter

        # TODO:: REMOVE - Verificar se existe filtro definido

        # Realiza busca na colecção para recuperar documentos que serão removidos
        find_result = collection.find(query_filter)
        find_result_count = find_result.count()

        # Verifica se existe documentos para ser removido
        if find_result_count > 0:
            # Recupera a coleção trash
            collection_trash = self.__collection_trash()
            # Grava os documentos localizados na coleção trash
            insert_many_result = collection_trash.insert_many(find_result)

            # Verifica se os documentos foram realmente gravados na coleção trash
            if len(insert_many_result.inserted_ids) == find_result_count:
                # Remove os documentos da coleção principal
                # Remove documento no banco de dados
                delete_result = collection.delete_many(query_filter)
                deleted_count = delete_result.deleted_count

        # Retorna informações das operaçoes realizadas
        conditions = {
            "query": {
                "filter": query_filter
            },
            "count": find_result_count
        }

        self.__datarest.result({"deleted_count": deleted_count}, conditions)

    # ----------------------------------------------------------------
    #
    # Private Methods
    #
    # ----------------------------------------------------------------
    # --------------------------------
    # __collection
    # --------------------------------
    # Recupera instancia da coleção que receberá interação na base de dados
    def __collection(self):
        # Recupera o banco de dados mongo db
        database = self.mongo_server[self.__datarest.database()]
        # Recupera/Cria a coleção mongo db
        return database[self.__datarest.collection()]

    def __collection_trash(self):
        # Recupera o banco de dados mongo db
        database = self.mongo_server[self.__datarest.database_name]
        # Recupera/Cria a coleção mongo db
        return database[self.__datarest.collection_name + "_trash"]

    def __collection_changes(self):
        # Recupera o banco de dados mongo db
        database = self.mongo_server[self.__datarest.database_name]
        # Recupera/Cria a coleção mongo db
        return database[self.__datarest.collection_name + "_changes"]

    # --------------------------------
    # __fields
    # --------------------------------
    def __fields(self, query_fields):
        list_fields = []

        for item_field in query_fields:
            if query_fields[item_field] == True:
                list_fields.append(item_field)
        return list_fields

    # --------------------------------
    # __prepare_result
    # --------------------------------
    def __prepare_result(self, data):
        def action(item_result):
            result = {}
            for item_field in item_result:
                # Converte ObjectId para string
                if type(item_result[item_field]) == bson.ObjectId:
                    result['id'] = str(item_result[item_field])

                # datetime.datetime
                # Converte datetime
                elif type(item_result[item_field]) == datetime.datetime:
                    format = "%a %b %d %H:%M:%S %Y"
                    result[item_field] = (item_result[item_field]).isoformat()

                # Converte Decimal128 para Float
                elif type(item_result[item_field]) == bson.decimal128.Decimal128:
                    result[item_field] = float(str(item_result[item_field]))

                # TODO:: Estudar datas do Python:  https://docs.python.org/2/library/datetime.html
                # TODO:: Extrair timestamp no horário correto
                # Converte Timestamp para number
                elif type(item_result[item_field]) == bson.timestamp.Timestamp:
                    result[item_field] = time.mktime((item_result[item_field]).as_datetime().timetuple())
                    # datetime.timestamp() === 1234567890   time.time()
                    # datetime.isoformat(sep='T', timespec='auto') === YYYY-MM-DDTHH:MM:SS.mmmmmm
                else:
                    result[item_field] = item_result[item_field]

            return result

        return list(map(action, data))
